import json

from vonage import Ncco
import ncco_samples.ncco_builder_samples as nbs


def test_build_basic_ncco():
    ncco = Ncco.build_ncco(nbs.talk_minimal)
    assert ncco == nbs.basic_ncco


def test_build_ncco_from_args():
    ncco = Ncco.build_ncco(nbs.record, nbs.talk_minimal)
    assert ncco == nbs.two_part_ncco
    assert (
        json.dumps(ncco)
        == '[{"action": "record", "eventUrl": ["http://example.com/events"]}, {"action": "talk", "text": "hello"}]'
    )


def test_build_ncco_from_list():
    action_list = [nbs.record, nbs.connect_advancedMachineDetection, nbs.talk_minimal]
    ncco = Ncco.build_ncco(actions=action_list)
    assert ncco == nbs.three_part_advancedMachineDetection_ncco


def test_build_insane_ncco():
    action_list = [
        nbs.record,
        nbs.conversation,
        nbs.connect,
        nbs.talk,
        nbs.stream,
        nbs.input,
        nbs.notify,
        nbs.get_pay_voice_prompt(),
        nbs.get_pay_text_prompt(),
    ]
    ncco = Ncco.build_ncco(actions=action_list)
    assert ncco == nbs.insane_ncco



from util import *
from vonage.errors import InvalidAuthenticationTypeError


@responses.activate
def test_get_with_query_params_auth(client, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/applications")
    host = "api.nexmo.com"
    request_uri = "/v1/applications"
    params = {"aaa": "xxx", "bbb": "yyy"}
    response = client.get(host, request_uri, params=params, auth_type='params')
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "aaa=xxx" in request_query()
    assert "bbb=yyy" in request_query()


@responses.activate
def test_get_with_header_auth(client, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/applications")
    host = "api.nexmo.com"
    request_uri = "/v1/applications"
    params = {"aaa": "xxx", "bbb": "yyy"}
    response = client.get(host, request_uri, params=params, auth_type='header')
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "aaa=xxx" in request_query()
    assert "bbb=yyy" in request_query()
    assert_basic_auth()


@responses.activate
def test_post_with_query_params_auth(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/v1/applications")
    host = "api.nexmo.com"
    request_uri = "/v1/applications"
    params = {"aaa": "xxx", "bbb": "yyy"}
    response = client.post(host, request_uri, params, auth_type='params', body_is_json=False)
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "aaa=xxx" in request_body()
    assert "bbb=yyy" in request_body()


@responses.activate
def test_post_with_header_auth(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/v1/applications")
    host = "api.nexmo.com"
    request_uri = "/v1/applications"
    params = {"aaa": "xxx", "bbb": "yyy"}
    response = client.post(host, request_uri, params, auth_type='header', body_is_json=False)
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "aaa=xxx" in request_body()
    assert "bbb=yyy" in request_body()
    assert_basic_auth()


@responses.activate
def test_put_with_header_auth(client, dummy_data):
    stub(responses.PUT, "https://api.nexmo.com/v1/applications")
    host = "api.nexmo.com"
    request_uri = "/v1/applications"
    params = {"aaa": "xxx", "bbb": "yyy"}
    response = client.put(host, request_uri, params=params, auth_type='header')
    assert_basic_auth()
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert b"aaa" in request_body()
    assert b"xxx" in request_body()
    assert b"bbb" in request_body()
    assert b"yyy" in request_body()


@responses.activate
def test_delete_with_header_auth(client, dummy_data):
    stub(responses.DELETE, "https://api.nexmo.com/v1/applications")
    host = "api.nexmo.com"
    request_uri = "/v1/applications"
    response = client.delete(host, request_uri, auth_type='header')
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert_basic_auth()


@responses.activate
def test_patch(client, dummy_data):
    stub(responses.PATCH, "https://api.nexmo.com/v1/applications")
    host = "api.nexmo.com"
    request_uri = "/v1/applications"
    params = {"aaa": "xxx", "bbb": "yyy"}
    response = client.patch(host, request_uri, params=params, auth_type='jwt')
    assert request_headers()['Content-Type'] == 'application/json'
    assert re.search(b'^Bearer ', request_headers()['Authorization']) is not None
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert b"aaa" in request_body()
    assert b"xxx" in request_body()
    assert b"bbb" in request_body()
    assert b"yyy" in request_body()


@responses.activate
def test_patch_no_content(client, dummy_data):
    stub(
        responses.PATCH,
        f"https://api.nexmo.com/v2/project",
        status_code=204,
        fixture_path='no_content.json',
    )
    host = "api.nexmo.com"
    request_uri = "/v2/project"
    params = {"test_param_1": "test1", "test_param_2": "test2"}
    client.patch(host, request_uri, params=params, auth_type='jwt')
    assert request_headers()['Content-Type'] == 'application/json'
    assert re.search(b'^Bearer ', request_headers()['Authorization']) is not None
    assert request_user_agent() == dummy_data.user_agent
    assert b"test_param_1" in request_body()
    assert b"test1" in request_body()
    assert b"test_param_2" in request_body()
    assert b"test2" in request_body()


def test_patch_invalid_auth_type(client):
    host = "api.nexmo.com"
    request_uri = "/v2/project"
    params = {"test_param_1": "test1", "test_param_2": "test2"}
    with pytest.raises(InvalidAuthenticationTypeError):
        client.patch(host, request_uri, params=params, auth_type='params')


@responses.activate
def test_get_with_jwt_auth(client, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/calls")
    host = "api.nexmo.com"
    request_uri = "/v1/calls"
    response = client.get(host, request_uri, auth_type='jwt')
    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent



from time import time
from unittest.mock import patch
from pytest import raises

from vonage import Client, ClientError

now = int(time())


def test_auth_sets_claims_from_kwargs(client):
    client.auth(jti='asdfzxcv1234', nbf=now + 100, exp=now + 1000)
    assert client._jwt_claims['jti'] == 'asdfzxcv1234'
    assert client._jwt_claims['nbf'] == now + 100
    assert client._jwt_claims['exp'] == now + 1000


def test_auth_sets_claims_from_dict(client):
    custom_jwt_claims = {'jti': 'asdfzxcv1234', 'nbf': now + 100, 'exp': now + 1000}
    client.auth(custom_jwt_claims)
    assert client._jwt_claims['jti'] == 'asdfzxcv1234'
    assert client._jwt_claims['nbf'] == now + 100
    assert client._jwt_claims['exp'] == now + 1000


test_jwt = b'eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJhcHBsaWNhdGlvbl9pZCI6ImFzZGYxMjM0IiwiaWF0IjoxNjg1NzMxMzkxLCJqdGkiOiIwYzE1MDJhZS05YmI5LTQ4YzQtYmQyZC0yOGFhNWUxYjZkMTkiLCJleHAiOjE2ODU3MzIyOTF9.mAkGeVgWOb7Mrzka7DSj32vSM8RaFpYse_2E7jCQ4DuH8i32wq9FxXGgfwdBQDHzgku3RYIjLM1xlVrGjNM3MsnZgR7ymQ6S4bdTTOmSK0dKbk91SrN7ZAC9k2a6JpCC2ZYgXpZ5BzpDTdy9BYu6msHKmkL79_aabFAhrH36Nk26pLvoI0-KiGImEex-aRR4iiaXhOebXBeqiQTRPKoKizREq4-8zBQv_j6yy4AiEYvBatQ8L_sjHsLj9jjITreX8WRvEW-G4TPpPLMaHACHTDMpJSOZAnegAkzTV2frVRmk6DyVXnemm4L0RQD1XZDaH7JPsKk24Hd2WZQyIgHOqQ'


def vonage_jwt_mock(self, claims):
    return test_jwt


def test_generate_application_jwt(client):
    with patch('vonage.client.JwtClient.generate_application_jwt', vonage_jwt_mock):
        jwt = client._generate_application_jwt()
    assert jwt == test_jwt


def test_create_jwt_auth_string(client):
    headers = client.headers
    with patch('vonage.client.JwtClient.generate_application_jwt', vonage_jwt_mock):
        headers['Authorization'] = client._create_jwt_auth_string()
    assert headers['Accept'] == 'application/json'
    assert headers['Authorization'] == b'Bearer ' + test_jwt


def test_create_jwt_error_no_application_id_or_private_key():
    empty_client = Client()

    with raises(ClientError) as err:
        empty_client._generate_application_jwt()
    assert (
        str(err.value)
        == 'JWT generation failed. Check that you passed in valid values for "application_id" and "private_key".'
    )



from __future__ import annotations
from typing import TYPE_CHECKING

from .camara_auth import CamaraAuth

if TYPE_CHECKING:
    from vonage import Client


class SimSwap:
    """Class containing methods for working with the Vonage SIM Swap API."""

    def __init__(self, client: Client):
        self._client = client
        self._auth_type = 'oauth2'
        self._camara_auth = CamaraAuth(client)

    def check(self, phone_number: str, max_age: int = None):
        """Check if a SIM swap has been performed in a given time frame.

        Args:
            phone_number (str): The phone number to check. Use the E.164 format without a leading +.
            max_age (int, optional): Period in hours to be checked for SIM swap.

        Returns:
            The response from the API as a dict.
        """
        oicd_response = self._camara_auth.make_oidc_request(
            number=phone_number, scope='dpv:FraudPreventionAndDetection#check-sim-swap'
        )
        token = self._camara_auth.request_backend_camara_token(oicd_response)

        params = {'phoneNumber': phone_number}
        if max_age:
            params['maxAge'] = max_age

        return self._client.post(
            'api-eu.vonage.com',
            '/camara/sim-swap/v040/check',
            params=params,
            auth_type=self._auth_type,
            oauth_token=token,
        )

    def get_last_swap_date(self, phone_number: str):
        """Get the last SIM swap date for a phone number.

        Args:
            phone_number (str): The phone number to check. Use the E.164 format without a leading +.

        Returns:
            The response from the API as a dict.
        """
        oicd_response = self._camara_auth.make_oidc_request(
            number=phone_number,
            scope='dpv:FraudPreventionAndDetection#retrieve-sim-swap-date',
        )
        token = self._camara_auth.request_backend_camara_token(oicd_response)
        return self._client.post(
            'api-eu.vonage.com',
            '/camara/sim-swap/v040/retrieve-date',
            params={'phoneNumber': phone_number},
            auth_type=self._auth_type,
            oauth_token=token,
        )



from util import *
from vonage import Client
from vonage.errors import (
    ClientError,
    VideoError,
    InvalidRoleError,
    TokenExpiryError,
    SipError,
)

import jwt
from time import time


session_id = 'my_session_id'
stream_id = 'my_stream_id'
connection_id = '1234-5678'
archive_id = '1234-abcd'
broadcast_id = '1748b7070a81464c9759c46ad10d3734'


@responses.activate
def test_create_default_session(client: Client, dummy_data):
    stub(
        responses.POST,
        "https://video.api.vonage.com/session/create",
        fixture_path="video/create_session.json",
    )

    session_info = client.video.create_session()
    assert isinstance(session_info, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert session_info['session_id'] == session_id
    assert session_info['archive_mode'] == 'manual'
    assert session_info['media_mode'] == 'routed'
    assert session_info['location'] == None


@responses.activate
def test_create_session_custom_archive_mode_and_location(client: Client):
    stub(
        responses.POST,
        "https://video.api.vonage.com/session/create",
        fixture_path="video/create_session.json",
    )

    session_options = {
        'archive_mode': 'always',
        'location': '192.0.1.1',
        'media_mode': 'routed',
    }
    session_info = client.video.create_session(session_options)
    assert isinstance(session_info, dict)
    assert session_info['session_id'] == session_id
    assert session_info['archive_mode'] == 'always'
    assert session_info['media_mode'] == 'routed'
    assert session_info['location'] == '192.0.1.1'


@responses.activate
def test_create_session_custom_media_mode(client: Client):
    stub(
        responses.POST,
        "https://video.api.vonage.com/session/create",
        fixture_path="video/create_session.json",
    )

    session_options = {'media_mode': 'relayed'}
    session_info = client.video.create_session(session_options)
    assert isinstance(session_info, dict)
    assert session_info['session_id'] == session_id
    assert session_info['archive_mode'] == 'manual'
    assert session_info['media_mode'] == 'relayed'
    assert session_info['location'] == None


def test_create_session_invalid_archive_mode(client: Client):
    session_options = {'archive_mode': 'invalid_option'}
    with pytest.raises(VideoError) as excinfo:
        client.video.create_session(session_options)
    assert 'Invalid archive_mode value. Must be one of ' in str(excinfo.value)


def test_create_session_invalid_media_mode(client: Client):
    session_options = {'media_mode': 'invalid_option'}
    with pytest.raises(VideoError) as excinfo:
        client.video.create_session(session_options)
    assert 'Invalid media_mode value. Must be one of ' in str(excinfo.value)


def test_create_session_invalid_mode_combination(client: Client):
    session_options = {'archive_mode': 'always', 'media_mode': 'relayed'}
    with pytest.raises(VideoError) as excinfo:
        client.video.create_session(session_options)
    assert (
        str(excinfo.value)
        == 'Invalid combination: cannot specify "archive_mode": "always" and "media_mode": "relayed".'
    )


def test_generate_client_token_all_defaults(client: Client):
    token = client.video.generate_client_token(session_id)
    decoded_token = jwt.decode(token, algorithms='RS256', options={'verify_signature': False})
    assert decoded_token['application_id'] == 'nexmo-application-id'
    assert decoded_token['scope'] == 'session.connect'
    assert decoded_token['session_id'] == 'my_session_id'
    assert decoded_token['role'] == 'publisher'
    assert decoded_token['initial_layout_class_list'] == ''


def test_generate_client_token_custom_options(client: Client):
    now = int(time())
    token_options = {
        'role': 'moderator',
        'data': 'some token data',
        'initialLayoutClassList': ['1234', '5678', '9123'],
        'expireTime': now + 60,
        'jti': 1234,
        'iat': now,
        'subject': 'test_subject',
        'acl': ['1', '2', '3'],
    }

    token = client.video.generate_client_token(session_id, token_options)
    decoded_token = jwt.decode(token, algorithms='RS256', options={'verify_signature': False})
    assert decoded_token['application_id'] == 'nexmo-application-id'
    assert decoded_token['scope'] == 'session.connect'
    assert decoded_token['session_id'] == 'my_session_id'
    assert decoded_token['role'] == 'moderator'
    assert decoded_token['initial_layout_class_list'] == ['1234', '5678', '9123']
    assert decoded_token['data'] == 'some token data'
    assert decoded_token['jti'] == 1234
    assert decoded_token['subject'] == 'test_subject'
    assert decoded_token['acl'] == ['1', '2', '3']


def test_generate_client_token_publisher_only_role(client: Client):
    token = client.video.generate_client_token(session_id, {'role': 'publisheronly'})
    decoded_token = jwt.decode(token, algorithms='RS256', options={'verify_signature': False})
    assert decoded_token['role'] == 'publisheronly'


def test_check_client_token_headers(client: Client):
    token = client.video.generate_client_token(session_id)
    headers = jwt.get_unverified_header(token)
    assert headers['alg'] == 'RS256'
    assert headers['typ'] == 'JWT'


def test_generate_client_token_invalid_role(client: Client):
    with pytest.raises(InvalidRoleError):
        client.video.generate_client_token(session_id, {'role': 'observer'})


def test_generate_client_token_invalid_expire_time(client: Client):
    now = int(time())
    with pytest.raises(TokenExpiryError):
        client.video.generate_client_token(session_id, {'expireTime': now + 3600 * 24 * 30 + 1})


@responses.activate
def test_get_stream(client: Client):
    stub(
        responses.GET,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/stream/{stream_id}",
        fixture_path="video/get_stream.json",
    )

    stream = client.video.get_stream(session_id, stream_id)
    assert isinstance(stream, dict)
    assert stream['videoType'] == 'camera'


@responses.activate
def test_list_streams(
    client: Client,
):
    stub(
        responses.GET,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/stream",
        fixture_path="video/list_streams.json",
    )

    stream_list = client.video.list_streams(session_id)
    assert isinstance(stream_list, dict)
    assert stream_list['items'][0]['videoType'] == 'camera'


@responses.activate
def test_change_stream_layout(client: Client):
    stub(
        responses.PUT,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/stream",
    )

    items = [{'id': 'stream-1234', 'layoutClassList': ["full"]}]

    assert isinstance(client.video.set_stream_layout(session_id, items), dict)
    assert request_content_type() == "application/json"


@responses.activate
def test_send_signal_to_all_participants(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/signal",
    )

    assert isinstance(
        client.video.send_signal(session_id, type='chat', data='hello from a test case'),
        dict,
    )
    assert request_content_type() == "application/json"


@responses.activate
def test_send_signal_to_single_participant(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/connection/{connection_id}/signal",
    )

    assert isinstance(
        client.video.send_signal(
            session_id,
            type='chat',
            data='hello from a test case',
            connection_id=connection_id,
        ),
        dict,
    )
    assert request_content_type() == "application/json"


@responses.activate
def test_disconnect_client(client: Client):
    stub(
        responses.DELETE,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/connection/{connection_id}",
    )

    assert isinstance(client.video.disconnect_client(session_id, connection_id=connection_id), dict)


@responses.activate
def test_mute_specific_stream(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/stream/{stream_id}/mute",
        fixture_path="video/mute_specific_stream.json",
    )

    response = client.video.mute_stream(session_id, stream_id)
    assert isinstance(response, dict)
    assert response['createdAt'] == 1414642898000


@responses.activate
def test_mute_all_streams(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/mute",
        fixture_path="video/mute_multiple_streams.json",
    )

    response = client.video.mute_all_streams(session_id)
    assert isinstance(response, dict)
    assert response['createdAt'] == 1414642898000


@responses.activate
def test_mute_all_streams_except_excluded_list(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/mute",
        fixture_path="video/mute_multiple_streams.json",
    )

    response = client.video.mute_all_streams(
        session_id, excluded_stream_ids=['excluded_stream_id_1', 'excluded_stream_id_2']
    )
    assert isinstance(response, dict)
    assert response['createdAt'] == 1414642898000


@responses.activate
def test_disable_mute_all_streams(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/mute",
        fixture_path="video/disable_mute_multiple_streams.json",
    )

    response = client.video.disable_mute_all_streams(
        session_id, excluded_stream_ids=['excluded_stream_id_1', 'excluded_stream_id_2']
    )
    assert isinstance(response, dict)
    assert (
        request_body()
        == b'{"active": false, "excludedStreamIds": ["excluded_stream_id_1", "excluded_stream_id_2"]}'
    )
    assert response['createdAt'] == 1414642898000


@responses.activate
def test_list_archives_with_filters_applied(client: Client):
    stub(
        responses.GET,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive",
        fixture_path="video/list_archives.json",
    )

    response = client.video.list_archives(offset=0, count=1, session_id=session_id)
    assert isinstance(response, dict)
    assert response['items'][0]['createdAt'] == 1384221730000
    assert response['items'][0]['streams'][0]['streamId'] == 'abc123'


@responses.activate
def test_create_new_archive(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive",
        fixture_path="video/create_archive.json",
    )

    response = client.video.create_archive(
        session_id=session_id, name='my_new_archive', outputMode='individual'
    )
    assert isinstance(response, dict)
    assert response['name'] == 'my_new_archive'
    assert response['createdAt'] == 1384221730555


@responses.activate
def test_get_archive(client: Client):
    stub(
        responses.GET,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive/{archive_id}",
        fixture_path="video/get_archive.json",
    )

    response = client.video.get_archive(archive_id=archive_id)
    assert isinstance(response, dict)
    assert response['duration'] == 5049
    assert response['size'] == 247748791
    assert response['streams'] == []


@responses.activate
def test_delete_archive(client: Client):
    stub(
        responses.GET,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive/{archive_id}",
        status_code=204,
        fixture_path='no_content.json',
    )

    assert client.video.delete_archive(archive_id=archive_id) == None


@responses.activate
def test_add_stream_to_archive(client: Client):
    stub(
        responses.PATCH,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive/{archive_id}/streams",
        status_code=204,
        fixture_path='no_content.json',
    )

    assert (
        client.video.add_stream_to_archive(
            archive_id=archive_id, stream_id='1234', has_audio=True, has_video=True
        )
        == None
    )


@responses.activate
def test_remove_stream_from_archive(client: Client):
    stub(
        responses.PATCH,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive/{archive_id}/streams",
        status_code=204,
        fixture_path='no_content.json',
    )

    assert client.video.remove_stream_from_archive(archive_id=archive_id, stream_id='1234') == None


@responses.activate
def test_stop_archive(client: Client):
    stub(
        responses.POST,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive/{archive_id}/stop",
        fixture_path="video/stop_archive.json",
    )

    response = client.video.stop_archive(archive_id=archive_id)
    assert response['name'] == 'my_new_archive'
    assert response['createdAt'] == 1384221730555
    assert response['status'] == 'stopped'


@responses.activate
def test_change_archive_layout(client: Client):
    stub(
        responses.PUT,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/archive/{archive_id}/layout",
    )

    params = {'type': 'bestFit', 'screenshareType': 'horizontalPresentation'}

    assert isinstance(client.video.change_archive_layout(archive_id, params), dict)
    assert request_content_type() == "application/json"


@responses.activate
def test_create_sip_call(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/dial',
        fixture_path='video/create_sip_call.json',
    )

    sip = {'uri': 'sip:user@sip.partner.com;transport=tls'}

    sip_call = client.video.create_sip_call(session_id, 'my_token', sip)
    assert sip_call['id'] == 'b0a5a8c7-dc38-459f-a48d-a7f2008da853'
    assert sip_call['connectionId'] == 'e9f8c166-6c67-440d-994a-04fb6dfed007'
    assert sip_call['streamId'] == '482bce73-f882-40fd-8ca5-cb74ff416036'


@responses.activate
def test_create_sip_call_not_found_error(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/dial',
        status_code=404,
    )
    sip = {'uri': 'sip:user@sip.partner.com;transport=tls'}
    with pytest.raises(ClientError):
        client.video.create_sip_call('an-invalid-session-id', 'my_token', sip)


def test_create_sip_call_no_uri_error(client):
    sip = {}
    with pytest.raises(SipError) as err:
        client.video.create_sip_call(session_id, 'my_token', sip)

    assert str(err.value) == 'You must specify a uri when creating a SIP call.'


@responses.activate
def test_play_dtmf(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/play-dtmf',
        fixture_path='no_content.json',
    )

    assert client.video.play_dtmf(session_id, '1234') == None


@responses.activate
def test_play_dtmf_specific_connection(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/connection/my-connection-id/play-dtmf',
        fixture_path='no_content.json',
    )

    assert client.video.play_dtmf(session_id, '1234', connection_id='my-connection-id') == None


@responses.activate
def test_play_dtmf_invalid_session_id_error(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/session/{session_id}/play-dtmf',
        fixture_path='video/play_dtmf_invalid_error.json',
        status_code=400,
    )

    with pytest.raises(ClientError) as err:
        client.video.play_dtmf(session_id, '1234')
    assert 'One of the properties digits or sessionId is invalid.' in str(err.value)


def test_play_dtmf_invalid_input_error(client):
    with pytest.raises(VideoError) as err:
        client.video.play_dtmf(session_id, '!@£$%^&()asdfghjkl;')

    assert str(err.value) == 'Only digits 0-9, *, #, and "p" are allowed.'


@responses.activate
def test_list_broadcasts(client):
    stub(
        responses.GET,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast',
        fixture_path='video/list_broadcasts.json',
    )

    broadcasts = client.video.list_broadcasts()
    assert broadcasts['count'] == '1'
    assert broadcasts['items'][0]['id'] == '1748b7070a81464c9759c46ad10d3734'
    assert broadcasts['items'][0]['applicationId'] == 'abc123'


@responses.activate
def test_list_broadcasts_options(client):
    stub(
        responses.GET,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast',
        fixture_path='video/list_broadcasts.json',
    )

    broadcasts = client.video.list_broadcasts(
        count=1, session_id='2_MX4xMDBfjE0Mzc2NzY1NDgwMTJ-TjMzfn4'
    )
    assert broadcasts['count'] == '1'
    assert broadcasts['items'][0]['sessionId'] == '2_MX4xMDBfjE0Mzc2NzY1NDgwMTJ-TjMzfn4'
    assert broadcasts['items'][0]['id'] == '1748b7070a81464c9759c46ad10d3734'
    assert broadcasts['items'][0]['applicationId'] == 'abc123'


@responses.activate
def test_list_broadcasts_invalid_options_errors(client):
    stub(
        responses.GET,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast',
        fixture_path='video/list_broadcasts.json',
    )

    with pytest.raises(VideoError) as err:
        client.video.list_broadcasts(offset=-2, session_id='2_MX4xMDBfjE0Mzc2NzY1NDgwMTJ-TjMzfn4')
    assert str(err.value) == 'Offset must be an int >= 0.'

    with pytest.raises(VideoError) as err:
        client.video.list_broadcasts(count=9999, session_id='2_MX4xMDBfjE0Mzc2NzY1NDgwMTJ-TjMzfn4')
    assert str(err.value) == 'Count must be an int between 0 and 1000.'

    with pytest.raises(VideoError) as err:
        client.video.list_broadcasts(offset='10', session_id='2_MX4xMDBfjE0Mzc2NzY1NDgwMTJ-TjMzfn4')
    assert str(err.value) == 'Offset must be an int >= 0.'


@responses.activate
def test_start_broadcast_required_params(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast',
        fixture_path='video/broadcast.json',
    )

    params = {
        "sessionId": "2_MX40NTMyODc3Mn5-fg",
        "outputs": {
            "rtmp": [
                {
                    "id": "foo",
                    "serverUrl": "rtmps://myfooserver/myfooapp",
                    "streamName": "myfoostream",
                }
            ]
        },
    }

    broadcast = client.video.start_broadcast(params)
    assert broadcast['id'] == '1748b7070a81464c9759c46ad10d3734'
    assert broadcast['createdAt'] == 1437676551000
    assert broadcast['maxBitrate'] == 2000000
    assert broadcast['broadcastUrls']['rtmp'][0]['id'] == 'abc123'


@responses.activate
def test_start_broadcast_all_params(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast',
        fixture_path='video/broadcast.json',
    )

    params = {
        "sessionId": "2_MX40NTMyODc3Mn5-fg",
        "layout": {
            "type": "custom",
            "stylesheet": "the layout stylesheet (only used with type == custom)",
            "screenshareType": "horizontalPresentation",
        },
        "maxDuration": 5400,
        "outputs": {
            "rtmp": [
                {
                    "id": "foo",
                    "serverUrl": "rtmps://myfooserver/myfooapp",
                    "streamName": "myfoostream",
                }
            ]
        },
        "resolution": "1920x1080",
        "streamMode": "manual",
        "multiBroadcastTag": "foo",
    }

    broadcast = client.video.start_broadcast(params)
    assert broadcast['id'] == '1748b7070a81464c9759c46ad10d3734'
    assert broadcast['createdAt'] == 1437676551000
    assert broadcast['maxBitrate'] == 2000000
    assert broadcast['broadcastUrls']['rtmp'][0]['id'] == 'abc123'


@responses.activate
def test_get_broadcast(client):
    stub(
        responses.GET,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast/{broadcast_id}',
        fixture_path='video/broadcast.json',
    )

    broadcast = client.video.get_broadcast(broadcast_id)
    assert broadcast['id'] == '1748b7070a81464c9759c46ad10d3734'
    assert broadcast['sessionId'] == '2_MX4xMDBfjE0Mzc2NzY1NDgwMTJ-TjMzfn4'
    assert broadcast['updatedAt'] == 1437676551000
    assert broadcast['resolution'] == '640x480'


@responses.activate
def test_stop_broadcast(client):
    stub(
        responses.POST,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast/{broadcast_id}',
        fixture_path='video/broadcast.json',
    )

    broadcast = client.video.stop_broadcast(broadcast_id)
    assert broadcast['id'] == '1748b7070a81464c9759c46ad10d3734'
    assert broadcast['sessionId'] == '2_MX4xMDBfjE0Mzc2NzY1NDgwMTJ-TjMzfn4'
    assert broadcast['updatedAt'] == 1437676551000
    assert broadcast['resolution'] == '640x480'


@responses.activate
def test_change_broadcast_layout(client):
    stub(
        responses.PUT,
        f'https://video.api.vonage.com/v2/project/{client.application_id}/broadcast/{broadcast_id}/layout',
        fixture_path='no_content.json',
    )

    params = {
        "type": "bestFit",
        "stylesheet": "stream.instructor {position: absolute; width: 100%;  height:50%;}",
        "screenshareType": "pip",
    }

    assert client.video.change_broadcast_layout(broadcast_id, params) == None


@responses.activate
def test_add_stream_to_broadcast(client: Client, dummy_data):
    stub(
        responses.PATCH,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/broadcast/{broadcast_id}/streams",
        status_code=204,
        fixture_path='no_content.json',
    )

    assert (
        client.video.add_stream_to_broadcast(
            broadcast_id=broadcast_id, stream_id='1234', has_audio=True, has_video=True
        )
        == None
    )
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_remove_stream_from_broadcast(client: Client, dummy_data):
    stub(
        responses.PATCH,
        f"https://video.api.vonage.com/v2/project/{client.application_id}/broadcast/{broadcast_id}/streams",
        status_code=204,
        fixture_path='no_content.json',
    )

    assert (
        client.video.remove_stream_from_broadcast(broadcast_id=broadcast_id, stream_id='1234')
        == None
    )
    assert request_user_agent() == dummy_data.user_agent



from util import *
from vonage.errors import MessagesError


def test_invalid_send_message_params_object(messages):
    with pytest.raises(MessagesError) as err:
        messages.send_message('hi')
    assert (
        str(err.value) == 'Parameters to the send_message method must be specified as a dictionary.'
    )


def test_invalid_message_channel(messages):
    with pytest.raises(MessagesError) as err:
        messages.send_message(
            {
                'channel': 'carrier_pigeon',
                'message_type': 'text',
                'to': '12345678',
                'from': 'vonage',
                'text': 'my important message',
            }
        )
    assert '"carrier_pigeon" is an invalid message channel.' in str(err.value)


def test_invalid_message_type(messages):
    with pytest.raises(MessagesError) as err:
        messages.send_message(
            {
                'channel': 'sms',
                'message_type': 'video',
                'to': '12345678',
                'from': 'vonage',
                'video': 'my_url.com',
            }
        )
    assert '"video" is not a valid message type for channel "sms".' in str(err.value)


def test_invalid_recipient_not_string(messages):
    with pytest.raises(MessagesError) as err:
        messages.send_message(
            {
                'channel': 'sms',
                'message_type': 'text',
                'to': 12345678,
                'from': 'vonage',
                'text': 'my important message',
            }
        )
    assert str(err.value) == 'Message recipient ("to=12345678") not in a valid format.'


def test_invalid_recipient_number(messages):
    with pytest.raises(MessagesError) as err:
        messages.send_message(
            {
                'channel': 'sms',
                'message_type': 'text',
                'to': '+441234567890',
                'from': 'vonage',
                'text': 'my important message',
            }
        )
    assert str(err.value) == 'Message recipient number ("to=+441234567890") not in a valid format.'


def test_invalid_messenger_recipient(messages):
    with pytest.raises(MessagesError) as err:
        messages.send_message(
            {
                'channel': 'messenger',
                'message_type': 'text',
                'to': '',
                'from': 'vonage',
                'text': 'my important message',
            }
        )
    assert str(err.value) == 'Message recipient ID ("to=") not in a valid format.'


def test_invalid_sender(messages):
    with pytest.raises(MessagesError) as err:
        messages.send_message(
            {
                'channel': 'sms',
                'message_type': 'text',
                'to': '441234567890',
                'from': 1234,
                'text': 'my important message',
            }
        )
    assert (
        str(err.value)
        == 'Message sender ("frm=1234") set incorrectly. Set a valid name or number for the sender.'
    )


def test_set_client_ref(messages):
    messages._check_valid_client_ref(
        {
            'channel': 'sms',
            'message_type': 'text',
            'to': '441234567890',
            'from': 'vonage',
            'text': 'my important message',
            'client_ref': 'my client reference',
        }
    )
    assert messages._client_ref == 'my client reference'


def test_invalid_client_ref(messages):
    with pytest.raises(MessagesError) as err:
        messages._check_valid_client_ref(
            {
                'channel': 'sms',
                'message_type': 'text',
                'to': '441234567890',
                'from': 'vonage',
                'text': 'my important message',
                'client_ref': 'my client reference that is, in fact, a small, but significant amount longer than the 100 character limit imposed at this present juncture.',
            }
        )
    assert str(err.value) == 'client_ref can be a maximum of 100 characters.'


def test_whatsapp_template(messages):
    messages.validate_send_message_input(
        {
            'channel': 'whatsapp',
            'message_type': 'template',
            'to': '4412345678912',
            'from': 'vonage',
            'template': {'name': 'namespace:mytemplate'},
            'whatsapp': {'policy': 'deterministic', 'locale': 'en-GB'},
        }
    )


def test_set_messenger_optional_attribute(messages):
    messages.validate_send_message_input(
        {
            'channel': 'messenger',
            'message_type': 'text',
            'to': 'user_messenger_id',
            'from': 'vonage',
            'text': 'my important message',
            'messenger': {'category': 'response', 'tag': 'ACCOUNT_UPDATE'},
        }
    )


def test_set_viber_service_optional_attribute(messages):
    messages.validate_send_message_input(
        {
            'channel': 'viber_service',
            'message_type': 'text',
            'to': '44123456789',
            'from': 'vonage',
            'text': 'my important message',
            'viber_service': {'category': 'transaction', 'ttl': 30, 'type': 'text'},
        }
    )


def test_viber_service_video(messages):
    messages.validate_send_message_input(
        {
            'channel': 'viber_service',
            'message_type': 'video',
            'to': '44123456789',
            'from': 'vonage',
            'video': {
                'url': 'https://example.com/video.mp4',
                'caption': 'Look at this video',
                'thumb_url': 'https://example.com/thumbnail.jpg',
            },
            'viber_service': {
                'category': 'transaction',
                'duration': '120',
                'ttl': 30,
                'type': 'string',
            },
        }
    )


def test_viber_service_file(messages):
    messages.validate_send_message_input(
        {
            'channel': 'viber_service',
            'message_type': 'file',
            'to': '44123456789',
            'from': 'vonage',
            'video': {'url': 'https://example.com/files', 'name': 'example.pdf'},
            'viber_service': {'category': 'transaction', 'ttl': 30, 'type': 'string'},
        }
    )


def test_viber_service_text_action_button(messages):
    messages.validate_send_message_input(
        {
            'channel': 'viber_service',
            'message_type': 'text',
            'to': '44123456789',
            'from': 'vonage',
            'text': 'my important message',
            'viber_service': {
                'category': 'transaction',
                'ttl': 30,
                'type': 'string',
                'action': {'url': 'https://example.com/page1.html', 'text': 'Find out more'},
            },
        }
    )


def test_viber_service_image_action_button(messages):
    messages.validate_send_message_input(
        {
            'channel': 'viber_service',
            'message_type': 'image',
            'to': '44123456789',
            'from': 'vonage',
            'image': {
                'url': 'https://example.com/image.jpg',
                'caption': 'Check out this new promotion',
            },
            'viber_service': {
                'category': 'transaction',
                'ttl': 30,
                'type': 'string',
                'action': {'url': 'https://example.com/page1.html', 'text': 'Find out more'},
            },
        }
    )


def test_incomplete_input(messages):
    with pytest.raises(MessagesError) as err:
        messages.validate_send_message_input(
            {
                'channel': 'viber_service',
                'message_type': 'text',
                'to': '44123456789',
                'from': 'vonage',
                'text': 'my important message',
            }
        )
    assert (
        str(err.value)
        == 'You must specify all required properties for message channel "viber_service".'
    )


def test_whatsapp_sticker_id(messages):
    messages.validate_send_message_input(
        {
            'channel': 'whatsapp',
            'message_type': 'sticker',
            'sticker': {'id': '13aaecab-2485-4255-a0a7-97a2be6906b9'},
            'to': '44123456789',
            'from': 'vonage',
        }
    )


def test_whatsapp_sticker_url(messages):
    messages.validate_send_message_input(
        {
            'channel': 'whatsapp',
            'message_type': 'sticker',
            'sticker': {'url': 'https://example.com/sticker1.webp'},
            'to': '44123456789',
            'from': 'vonage',
        }
    )


def test_whatsapp_sticker_invalid_input_error(messages):
    with pytest.raises(MessagesError) as err:
        messages.validate_send_message_input(
            {
                'channel': 'whatsapp',
                'message_type': 'sticker',
                'sticker': {'my_sticker'},
                'to': '44123456789',
                'from': 'vonage',
            }
        )
    assert (
        str(err.value) == 'Must specify one, and only one, of "id" or "url" in the "sticker" field.'
    )


def test_whatsapp_sticker_exclusive_keys_error(messages):
    with pytest.raises(MessagesError) as err:
        messages.validate_send_message_input(
            {
                'channel': 'whatsapp',
                'message_type': 'sticker',
                'sticker': {
                    'id': '13aaecab-2485-4255-a0a7-97a2be6906b9',
                    'url': 'https://example.com/sticker1.webp',
                },
                'to': '44123456789',
                'from': 'vonage',
            }
        )
    assert (
        str(err.value) == 'Must specify one, and only one, of "id" or "url" in the "sticker" field.'
    )



import pytest
from vonage import Ncco, ConnectEndpoints, InputTypes, PayPrompts

record = Ncco.Record(eventUrl='http://example.com/events')

conversation = Ncco.Conversation(name='my_conversation')

connect = Ncco.Connect(
    endpoint=ConnectEndpoints.PhoneEndpoint(number='447000000000'),
    from_='447400000000',
    randomFromNumber=False,
    eventType='synchronous',
    timeout=15,
    limit=1000,
    machineDetection='hangup',
    eventUrl='http://example.com',
    eventMethod='PUT',
    ringbackTone='http://example.com',
)

connect_advancedMachineDetection = Ncco.Connect(
    endpoint=ConnectEndpoints.PhoneEndpoint(number='447000000000'),
    advancedMachineDetection={'behavior': 'continue', 'mode': 'detect'},
)


talk_minimal = Ncco.Talk(text='hello')

talk = Ncco.Talk(
    text='hello', bargeIn=True, loop=3, level=0.5, language='en-GB', style=1, premium=True
)

stream = Ncco.Stream(
    streamUrl='https://example.com/stream/music.mp3', level=0.1, bargeIn=True, loop=10
)

input = Ncco.Input(
    type=['dtmf', 'speech'],
    dtmf=InputTypes.Dtmf(timeOut=5, maxDigits=12, submitOnHash=True),
    speech=InputTypes.Speech(
        uuid='my-uuid',
        endOnSilence=2.5,
        language='en-GB',
        context=['sales', 'billing'],
        startTimeout=20,
        maxDuration=30,
        saveAudio=True,
    ),
    eventUrl='http://example.com/speech',
    eventMethod='put',
)

notify = Ncco.Notify(
    payload={"message": "world"}, eventUrl=["http://example.com"], eventMethod='PUT'
)


def get_pay_voice_prompt():
    with pytest.deprecated_call():
        return Ncco.Pay(
            amount=99.99,
            currency='gbp',
            eventUrl='https://example.com/payment',
            voice=PayPrompts.VoicePrompt(language='en-GB', style=1),
        )


def get_pay_text_prompt():
    with pytest.deprecated_call():
        return Ncco.Pay(
            amount=12.345,
            currency='gbp',
            eventUrl='https://example.com/payment',
            prompts=PayPrompts.TextPrompt(
                type='CardNumber',
                text='Enter your card number.',
                errors={
                    'InvalidCardType': {
                        'text': 'The card you are trying to use is not valid for this purchase.'
                    }
                },
            ),
        )


basic_ncco = [{"action": "talk", "text": "hello"}]

two_part_ncco = [
    {
        'action': 'record',
        'eventUrl': ['http://example.com/events'],
    },
    {'action': 'talk', 'text': 'hello'},
]

three_part_advancedMachineDetection_ncco = [
    {'action': 'record', 'eventUrl': ['http://example.com/events']},
    {
        'action': 'connect',
        'endpoint': [{'type': 'phone', 'number': '447000000000'}],
        'advancedMachineDetection': {'behavior': 'continue', 'mode': 'detect'},
    },
    {'action': 'talk', 'text': 'hello'},
]

insane_ncco = [
    {'action': 'record', 'eventUrl': ['http://example.com/events']},
    {'action': 'conversation', 'name': 'my_conversation'},
    {
        'action': 'connect',
        'endpoint': [{'number': '447000000000', 'type': 'phone'}],
        'eventMethod': 'PUT',
        'eventType': 'synchronous',
        'eventUrl': ['http://example.com'],
        'from': '447400000000',
        'limit': 1000,
        'machineDetection': 'hangup',
        'randomFromNumber': False,
        'ringbackTone': 'http://example.com',
        'timeout': 15,
    },
    {
        'action': 'talk',
        'bargeIn': True,
        'language': 'en-GB',
        'level': 0.5,
        'loop': 3,
        'premium': True,
        'style': 1,
        'text': 'hello',
    },
    {
        'action': 'stream',
        'bargeIn': True,
        'level': 0.1,
        'loop': 10,
        'streamUrl': ['https://example.com/stream/music.mp3'],
    },
    {
        'action': 'input',
        'dtmf': {'maxDigits': 12, 'submitOnHash': True, 'timeOut': 5},
        'eventMethod': 'PUT',
        'eventUrl': ['http://example.com/speech'],
        'speech': {
            'context': ['sales', 'billing'],
            'endOnSilence': 2.5,
            'language': 'en-GB',
            'maxDuration': 30,
            'saveAudio': True,
            'startTimeout': 20,
            'uuid': 'my-uuid',
        },
        'type': ['dtmf', 'speech'],
    },
    {
        'action': 'notify',
        'eventMethod': 'PUT',
        'eventUrl': ['http://example.com'],
        'payload': {'message': 'world'},
    },
    {
        'action': 'pay',
        'amount': 99.99,
        'currency': 'gbp',
        'eventUrl': ['https://example.com/payment'],
        'voice': {'language': 'en-GB', 'style': 1},
    },
    {
        'action': 'pay',
        'amount': 12.35,
        'currency': 'gbp',
        'eventUrl': ['https://example.com/payment'],
        'prompts': {
            'errors': {
                'InvalidCardType': {
                    'text': 'The card you are trying ' 'to use is not valid for ' 'this purchase.'
                }
            },
            'text': 'Enter your card number.',
            'type': 'CardNumber',
        },
    },
]



from pydantic import BaseModel, confloat, conint
from typing import List


class InputTypes:
    class Dtmf(BaseModel):
        timeOut: conint(ge=0, le=10) = None
        maxDigits: conint(ge=1, le=20) = None
        submitOnHash: bool = None

    class Speech(BaseModel):
        uuid: str = None
        endOnSilence: confloat(ge=0.4, le=10.0) = None
        language: str = None
        context: List[str] = None
        startTimeout: conint(ge=1, le=60) = None
        maxDuration: conint(ge=1, le=60) = None
        saveAudio: bool = None

    @classmethod
    def create_dtmf_model(cls, dict) -> Dtmf:
        return cls.Dtmf.model_validate(dict)

    @classmethod
    def create_speech_model(cls, dict) -> Speech:
        return cls.Speech.model_validate(dict)



import os


def test_subdirectories_are_python_packages():
    subdirs = [
        os.path.join('src/vonage', o)
        for o in os.listdir('src/vonage')
        if os.path.isdir(os.path.join('src/vonage', o))
    ]
    for subdir in subdirs:
        if '__pycache__' in subdir or os.path.isfile(f'{subdir}/__init__.py'):
            continue
        else:
            raise Exception(f'Subfolder {subdir} doesn\'t have an __init__.py file')



import vonage
from util import *


def test_check_signature(dummy_data):
    params = {
        "a": "1",
        "b": "2",
        "timestamp": "1461605396",
        "sig": "6af838ef94998832dbfc29020b564830",
    }

    client = vonage.Client(
        key=dummy_data.api_key, secret=dummy_data.api_secret, signature_secret="secret"
    )

    assert client.check_signature(params)


def test_signature(client, dummy_data):
    params = {"a": "1", "b": "2", "timestamp": "1461605396"}
    client = vonage.Client(
        key=dummy_data.api_key, secret=dummy_data.api_secret, signature_secret="secret"
    )
    assert client.signature(params) == "6af838ef94998832dbfc29020b564830"


def test_signature_adds_timestamp(dummy_data):
    params = {"a=7": "1", "b": "2 & 5"}

    client = vonage.Client(
        key=dummy_data.api_key, secret=dummy_data.api_secret, signature_secret="secret"
    )

    client.signature(params)
    assert params["timestamp"] is not None


def test_signature_md5(dummy_data):
    params = {"a": "1", "b": "2", "timestamp": "1461605396"}
    client = vonage.Client(
        key=dummy_data.api_key,
        secret=dummy_data.api_secret,
        signature_secret=dummy_data.signature_secret,
        signature_method="md5",
    )
    assert client.signature(params) == "c15c21ced558c93a226c305f58f902f2"


def test_signature_sha1(dummy_data):
    params = {"a": "1", "b": "2", "timestamp": "1461605396"}
    client = vonage.Client(
        key=dummy_data.api_key,
        secret=dummy_data.api_secret,
        signature_secret=dummy_data.signature_secret,
        signature_method="sha1",
    )
    assert client.signature(params) == "3e19a4e6880fdc2c1426bfd0587c98b9532f0210"


def test_signature_sha256(dummy_data):
    params = {"a": "1", "b": "2", "timestamp": "1461605396"}
    client = vonage.Client(
        key=dummy_data.api_key,
        secret=dummy_data.api_secret,
        signature_secret=dummy_data.signature_secret,
        signature_method="sha256",
    )
    assert (
        client.signature(params)
        == "a321e824b9b816be7c3f28859a31749a098713d39f613c80d455bbaffae1cd24"
    )


def test_signature_sha512(dummy_data):
    params = {"a": "1", "b": "2", "timestamp": "1461605396"}
    client = vonage.Client(
        key=dummy_data.api_key,
        secret=dummy_data.api_secret,
        signature_secret=dummy_data.signature_secret,
        signature_method="sha512",
    )
    assert (
        client.signature(params)
        == "812a18f76680fa0fe1b8bd9ee1625466ceb1bd96242e4d050d2cfd9a7b40166c63ed26ec9702168781b6edcf1633db8ff95af9341701004eec3fcf9550572ee8"
    )



from vonage import Client, ClientError, SubaccountsError
from util import stub

from pytest import raises
import responses

api_key = '1234asdf'
api_secret = 'qwerasdfzxcv'
client = Client(key=api_key, secret=api_secret)
subaccount_key = 'asdfzxcv'


@responses.activate
def test_list_subaccounts():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/list_subaccounts.json',
    )
    subaccounts = client.subaccounts.list_subaccounts()
    assert subaccounts['total_balance'] == 9.9999
    assert subaccounts['_embedded']['primary_account']['api_key'] == api_key
    assert subaccounts['_embedded']['primary_account']['balance'] == 9.9999
    assert subaccounts['_embedded']['subaccounts'][0]['api_key'] == 'qwerasdf'
    assert subaccounts['_embedded']['subaccounts'][0]['name'] == 'test_subaccount'


@responses.activate
def test_list_subaccounts_error_authentication():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )
    with raises(ClientError) as err:
        client.subaccounts.list_subaccounts()
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_list_subaccounts_error_forbidden():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/forbidden.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.list_subaccounts()
    assert (
        str(err.value)
        == 'Authorisation error: Account 1234adsf is not provisioned to access Subaccount Provisioning API (https://developer.nexmo.com/api-errors#unprovisioned)'
    )


@responses.activate
def test_list_subaccounts_error_not_found():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )
    with raises(ClientError) as err:
        client.subaccounts.list_subaccounts()
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


@responses.activate
def test_create_subaccount():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/subaccount.json',
    )
    subaccount = client.subaccounts.create_subaccount(
        name='my subaccount', secret='Password123', use_primary_account_balance=True
    )
    assert subaccount['api_key'] == 'asdfzxcv'
    assert subaccount['secret'] == 'Password123'
    assert subaccount['primary_account_api_key'] == api_key
    assert subaccount['use_primary_account_balance'] == True


@responses.activate
def test_create_subaccount_error_authentication():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )

    with raises(ClientError) as err:
        client.subaccounts.create_subaccount('failed subaccount')
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_create_subaccount_error_forbidden():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/forbidden.json',
        status_code=403,
    )

    with raises(ClientError) as err:
        client.subaccounts.create_subaccount('failed subaccount')
    assert (
        str(err.value)
        == 'Authorisation error: Account 1234adsf is not provisioned to access Subaccount Provisioning API (https://developer.nexmo.com/api-errors#unprovisioned)'
    )


@responses.activate
def test_create_subaccount_error_not_found():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        client.subaccounts.create_subaccount('failed subaccount')
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


def test_create_subaccount_error_non_boolean():
    with raises(SubaccountsError) as err:
        client.subaccounts.create_subaccount(
            'failed subaccount', use_primary_account_balance='yes please'
        )
    assert (
        str(err.value)
        == 'If providing a value, it needs to be a boolean. You provided: "yes please"'
    )


@responses.activate
def test_get_subaccount():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/subaccount.json',
    )
    subaccount = client.subaccounts.get_subaccount(subaccount_key)
    assert subaccount['api_key'] == 'asdfzxcv'
    assert subaccount['secret'] == 'Password123'
    assert subaccount['primary_account_api_key'] == api_key
    assert subaccount['use_primary_account_balance'] == True


@responses.activate
def test_get_subaccount_error_authentication():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )
    with raises(ClientError) as err:
        client.subaccounts.get_subaccount(subaccount_key)
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_get_subaccount_error_forbidden():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/forbidden.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.get_subaccount(subaccount_key)
    assert (
        str(err.value)
        == 'Authorisation error: Account 1234adsf is not provisioned to access Subaccount Provisioning API (https://developer.nexmo.com/api-errors#unprovisioned)'
    )


@responses.activate
def test_get_subaccount_error_not_found():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )
    with raises(ClientError) as err:
        client.subaccounts.get_subaccount(subaccount_key)
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


@responses.activate
def test_modify_subaccount():
    stub(
        responses.PATCH,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/modified_subaccount.json',
    )
    subaccount = client.subaccounts.modify_subaccount(
        subaccount_key,
        suspended=True,
        use_primary_account_balance=False,
        name='my modified subaccount',
    )
    assert subaccount['api_key'] == 'asdfzxcv'
    assert subaccount['name'] == 'my modified subaccount'
    assert subaccount['suspended'] == True
    assert subaccount['primary_account_api_key'] == api_key
    assert subaccount['use_primary_account_balance'] == False


@responses.activate
def test_modify_subaccount_error_authentication():
    stub(
        responses.PATCH,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )
    with raises(ClientError) as err:
        client.subaccounts.modify_subaccount(subaccount_key, suspended=True)
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_modify_subaccount_error_forbidden():
    stub(
        responses.PATCH,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/forbidden.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.modify_subaccount(subaccount_key, use_primary_account_balance=False)
    assert (
        str(err.value)
        == 'Authorisation error: Account 1234adsf is not provisioned to access Subaccount Provisioning API (https://developer.nexmo.com/api-errors#unprovisioned)'
    )


@responses.activate
def test_modify_subaccount_error_not_found():
    stub(
        responses.PATCH,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )
    with raises(ClientError) as err:
        client.subaccounts.modify_subaccount(subaccount_key, name='my modified subaccount name')
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


@responses.activate
def test_modify_subaccount_validation_error():
    stub(
        responses.PATCH,
        f'https://api.nexmo.com/accounts/{api_key}/subaccounts/{subaccount_key}',
        fixture_path='subaccounts/validation_error.json',
        status_code=422,
    )
    with raises(ClientError) as err:
        client.subaccounts.modify_subaccount(subaccount_key, use_primary_account_balance=True)
    assert (
        str(err.value)
        == 'Bad Request: The request failed due to validation errors (https://developer.nexmo.com/api-errors/account/subaccounts#validation)'
    )


@responses.activate
def test_list_credit_transfers():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/list_credit_transfers.json',
    )
    transfers = client.subaccounts.list_credit_transfers(
        start_date='2022-03-29T14:16:56Z',
        end_date='2023-06-12T17:20:01Z',
        subaccount='asdfzxcv',
    )
    assert transfers['_embedded']['credit_transfers'][0]['from'] == '1234asdf'
    assert transfers['_embedded']['credit_transfers'][0]['reference'] == 'test credit transfer'
    assert transfers['_embedded']['credit_transfers'][1]['to'] == 'asdfzxcv'
    assert transfers['_embedded']['credit_transfers'][1]['created_at'] == '2023-06-12T17:20:01.000Z'


@responses.activate
def test_list_credit_transfers_error_authentication():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )

    with raises(ClientError) as err:
        client.subaccounts.list_credit_transfers()
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_list_credit_transfers_error_forbidden():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/forbidden.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.list_credit_transfers()
    assert (
        str(err.value)
        == 'Authorisation error: Account 1234adsf is not provisioned to access Subaccount Provisioning API (https://developer.nexmo.com/api-errors#unprovisioned)'
    )


@responses.activate
def test_list_credit_transfers_error_not_found():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        client.subaccounts.list_credit_transfers()
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


@responses.activate
def test_list_credit_transfers_validation_error():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/transfer_validation_error.json',
        status_code=422,
    )
    with raises(ClientError) as err:
        client.subaccounts.list_credit_transfers(start_date='invalid-date-format')
    assert (
        str(err.value)
        == 'Bad Request: The request failed due to validation errors (https://developer.nexmo.com/api-errors/account/subaccounts#validation)'
    )


@responses.activate
def test_transfer_credit():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/credit_transfer.json',
    )
    transfer = client.subaccounts.transfer_credit(
        from_='1234asdf', to='asdfzxcv', amount=0.50, reference='test credit transfer'
    )
    assert transfer['from'] == '1234asdf'
    assert transfer['to'] == 'asdfzxcv'
    assert transfer['amount'] == 0.5
    assert transfer['reference'] == 'test credit transfer'


@responses.activate
def test_transfer_credit_error_authentication():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )

    with raises(ClientError) as err:
        client.subaccounts.transfer_credit(from_='1234asdf', to='asdfzxcv', amount=0.1)
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_transfer_credit_invalid_transfer():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/invalid_transfer.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.transfer_credit(from_='asdfzxcv', to='qwerasdf', amount=1)
    assert (
        str(err.value)
        == 'Invalid Transfer: Transfers are only allowed between a primary account and its subaccount (https://developer.nexmo.com/api-errors/account/subaccounts#valid-transfers)'
    )


@responses.activate
def test_transfer_credit_insufficient_credit():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/insufficient_credit.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.transfer_credit(from_='asdfzxcv', to='qwerasdf', amount=1)
    assert (
        str(err.value)
        == 'Transfer amount is invalid: Insufficient Credit (https://developer.nexmo.com/api-errors/account/subaccounts#valid-transfers)'
    )


@responses.activate
def test_transfer_credit_error_not_found():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        client.subaccounts.transfer_credit(from_='1234asdf', to='asdfzcv', amount=0.1)
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


@responses.activate
def test_transfer_credit_validation_error():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/credit-transfers',
        fixture_path='subaccounts/must_be_number.json',
        status_code=422,
    )
    with raises(ClientError) as err:
        client.subaccounts.transfer_credit(from_='1234asdf', to='asdfzxcv', amount='0.50')
    assert (
        str(err.value)
        == 'Bad Request: The request failed due to validation errors (https://developer.nexmo.com/api-errors/account/subaccounts#validation)'
    )


@responses.activate
def test_list_balance_transfers():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/list_balance_transfers.json',
    )
    transfers = client.subaccounts.list_balance_transfers(
        start_date='2022-03-29T14:16:56Z',
        end_date='2023-06-12T17:20:01Z',
        subaccount='asdfzxcv',
    )
    assert transfers['_embedded']['balance_transfers'][0]['from'] == '1234asdf'
    assert transfers['_embedded']['balance_transfers'][0]['reference'] == 'test transfer'
    assert transfers['_embedded']['balance_transfers'][1]['to'] == 'asdfzxcv'
    assert (
        transfers['_embedded']['balance_transfers'][1]['created_at'] == '2023-06-12T17:20:01.000Z'
    )


@responses.activate
def test_list_balance_transfers_error_authentication():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )

    with raises(ClientError) as err:
        client.subaccounts.list_balance_transfers()
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_list_balance_transfers_error_forbidden():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/forbidden.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.list_balance_transfers()
    assert (
        str(err.value)
        == 'Authorisation error: Account 1234adsf is not provisioned to access Subaccount Provisioning API (https://developer.nexmo.com/api-errors#unprovisioned)'
    )


@responses.activate
def test_list_balance_transfers_error_not_found():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        client.subaccounts.list_balance_transfers()
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


@responses.activate
def test_list_balance_transfers_validation_error():
    stub(
        responses.GET,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/transfer_validation_error.json',
        status_code=422,
    )
    with raises(ClientError) as err:
        client.subaccounts.list_balance_transfers(start_date='invalid-date-format')
    assert (
        str(err.value)
        == 'Bad Request: The request failed due to validation errors (https://developer.nexmo.com/api-errors/account/subaccounts#validation)'
    )


@responses.activate
def test_transfer_balance():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/balance_transfer.json',
    )
    transfer = client.subaccounts.transfer_balance(
        from_='1234asdf', to='asdfzxcv', amount=0.50, reference='test balance transfer'
    )
    assert transfer['from'] == '1234asdf'
    assert transfer['to'] == 'asdfzxcv'
    assert transfer['amount'] == 0.5
    assert transfer['reference'] == 'test balance transfer'


@responses.activate
def test_transfer_balance_error_authentication():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )

    with raises(ClientError) as err:
        client.subaccounts.transfer_balance(from_='1234asdf', to='asdfzxcv', amount=0.1)
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_transfer_balance_invalid_transfer():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/invalid_transfer.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.transfer_balance(from_='asdfzxcv', to='qwerasdf', amount=1)
    assert (
        str(err.value)
        == 'Invalid Transfer: Transfers are only allowed between a primary account and its subaccount (https://developer.nexmo.com/api-errors/account/subaccounts#valid-transfers)'
    )


@responses.activate
def test_transfer_balance_error_not_found():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        client.subaccounts.transfer_balance(from_='1234asdf', to='asdfzcv', amount=0.1)
    assert (
        str(err.value)
        == "Invalid API Key: API key '1234asdf' does not exist, or you do not have access (https://developer.nexmo.com/api-errors#invalid-api-key)"
    )


@responses.activate
def test_transfer_balance_validation_error():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/balance-transfers',
        fixture_path='subaccounts/must_be_number.json',
        status_code=422,
    )
    with raises(ClientError) as err:
        client.subaccounts.transfer_balance(from_='1234asdf', to='asdfzxcv', amount='0.50')
    assert (
        str(err.value)
        == 'Bad Request: The request failed due to validation errors (https://developer.nexmo.com/api-errors/account/subaccounts#validation)'
    )


@responses.activate
def test_transfer_number():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/transfer-number',
        fixture_path='subaccounts/transfer_number.json',
    )
    transfer = client.subaccounts.transfer_number(
        from_='1234asdf', to='asdfzxcv', number='12345678901', country='US'
    )
    assert transfer['from'] == '1234asdf'
    assert transfer['to'] == 'asdfzxcv'
    assert transfer['number'] == '12345678901'
    assert transfer['country'] == 'US'
    assert transfer['masterAccountId'] == '1234asdf'


@responses.activate
def test_transfer_number_error_authentication():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/transfer-number',
        fixture_path='subaccounts/invalid_credentials.json',
        status_code=401,
    )

    with raises(ClientError) as err:
        client.subaccounts.transfer_number(
            from_='1234asdf', to='asdfzxcv', number='12345678901', country='US'
        )
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_transfer_number_invalid_transfer():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/transfer-number',
        fixture_path='subaccounts/invalid_number_transfer.json',
        status_code=403,
    )
    with raises(ClientError) as err:
        client.subaccounts.transfer_number(
            from_='1234asdf', to='asdfzxcv', number='12345678901', country='US'
        )
    assert (
        str(err.value)
        == 'Invalid Number Transfer: Could not transfer number 12345678901 from account 1234asdf to asdfzxcv - ShortCode is not owned by from account (https://developer.nexmo.com/api-errors/account/subaccounts#invalid-number-transfer)'
    )


@responses.activate
def test_transfer_number_error_number_not_found():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/transfer-number',
        fixture_path='subaccounts/number_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        client.subaccounts.transfer_number(
            from_='1234asdf', to='asdfzxcv', number='12345678901', country='US'
        )
    assert (
        str(err.value)
        == 'Invalid Number Transfer: Could not transfer number 12345678901 from account 1234asdf to asdfzxcv - ShortCode not found (https://developer.nexmo.com/api-errors/account/subaccounts#missing-number-transfer)'
    )


@responses.activate
def test_transfer_number_error_number_not_found():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/transfer-number',
        fixture_path='subaccounts/number_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        client.subaccounts.transfer_number(
            from_='1234asdf', to='asdfzxcv', number='12345678901', country='US'
        )
    assert (
        str(err.value)
        == 'Invalid Number Transfer: Could not transfer number 12345678901 from account 1234asdf to asdfzxcv - ShortCode not found (https://developer.nexmo.com/api-errors/account/subaccounts#missing-number-transfer)'
    )


@responses.activate
def test_transfer_number_validation_error():
    stub(
        responses.POST,
        f'https://api.nexmo.com/accounts/{api_key}/transfer-number',
        fixture_path='subaccounts/same_from_and_to_accounts.json',
        status_code=422,
    )
    with raises(ClientError) as err:
        client.subaccounts.transfer_number(
            from_='asdfzxcv', to='asdfzxcv', number='12345678901', country='US'
        )
    assert (
        str(err.value)
        == 'Bad Request: The request failed due to validation errors (https://developer.nexmo.com/api-errors/account/subaccounts#validation)'
    )



from vonage.sim_swap import SimSwap
from util import *

import responses


@responses.activate
def test_check_sim_swap(sim_swap: SimSwap):
    stub(
        responses.POST,
        'https://api-eu.vonage.com/oauth2/bc-authorize',
        fixture_path='camara_auth/oidc_request.json',
    )
    stub(
        responses.POST,
        'https://api-eu.vonage.com/oauth2/token',
        fixture_path='camara_auth/token_request.json',
    )
    stub(
        responses.POST,
        'https://api-eu.vonage.com/camara/sim-swap/v040/check',
        fixture_path='sim_swap/check_sim_swap.json',
    )

    response = sim_swap.check('447700900000', max_age=24)

    assert response['swapped'] == True


@responses.activate
def test_get_last_swap_date(sim_swap: SimSwap):
    stub(
        responses.POST,
        'https://api-eu.vonage.com/oauth2/bc-authorize',
        fixture_path='camara_auth/oidc_request.json',
    )
    stub(
        responses.POST,
        'https://api-eu.vonage.com/oauth2/token',
        fixture_path='camara_auth/token_request.json',
    )
    stub(
        responses.POST,
        'https://api-eu.vonage.com/camara/sim-swap/v040/retrieve-date',
        fixture_path='sim_swap/get_swap_date.json',
    )

    response = sim_swap.get_last_swap_date('447700900000')

    assert response['latestSimChange'] == '2019-08-24T14:15:22Z'



from vonage import Client, Users
from util import *
from vonage.errors import UsersError, ClientError, ServerError

from pytest import raises
import responses

client = Client()
users = Users(client)
host = client.api_host()


@responses.activate
def test_list_users_basic():
    stub(
        responses.GET,
        f'https://{host}/v1/users',
        fixture_path='users/list_users_basic.json',
    )

    all_users = users.list_users()
    assert all_users['page_size'] == 10
    assert all_users['_embedded']['users'][0]['name'] == 'NAM-6dd4ea1f-3841-47cb-a3d3-e271f5c1e33c'
    assert all_users['_embedded']['users'][1]['name'] == 'NAM-ecb938f2-13e0-40c1-9d3b-b16ebb4ef3d1'
    assert all_users['_embedded']['users'][2]['name'] == 'my_user_name'


@responses.activate
def test_list_users_options():
    stub(
        responses.GET,
        f'https://{host}/v1/users',
        fixture_path='users/list_users_options.json',
    )

    all_users = users.list_users(page_size=2, order='desc')
    assert all_users['page_size'] == 2
    assert all_users['_embedded']['users'][0]['name'] == 'my_user_name'
    assert all_users['_embedded']['users'][1]['name'] == 'NAM-ecb938f2-13e0-40c1-9d3b-b16ebb4ef3d1'


def test_list_users_order_error():
    with raises(UsersError) as err:
        users.list_users(order='Why, ascending of course!')
    assert (
        str(err.value) == 'Invalid order parameter. Must be one of: "asc", "desc", "ASC", "DESC".'
    )


@responses.activate
def test_list_users_400():
    stub(
        responses.GET,
        f'https://{host}/v1/users',
        fixture_path='users/list_users_400.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        users.list_users(page_size='asdf')
    assert 'Input validation failure.' in str(err.value)


@responses.activate
def test_list_users_404():
    stub(
        responses.GET,
        f'https://{host}/v1/users',
        fixture_path='users/list_users_404.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        users.list_users(name='asdf')
    assert 'User does not exist, or you do not have access.' in str(err.value)


@responses.activate
def test_list_users_429():
    stub(
        responses.GET,
        f'https://{host}/v1/users',
        fixture_path='users/rate_limit.json',
        status_code=429,
    )

    with raises(ClientError) as err:
        users.list_users()
    assert 'You have exceeded your request limit. You can try again shortly.' in str(err.value)


@responses.activate
def test_list_users_500():
    stub(
        responses.GET,
        f'https://{host}/v1/users',
        fixture_path='users/list_users_500.json',
        status_code=500,
    )

    with raises(ServerError) as err:
        users.list_users()
    assert str(err.value) == '500 response from api.nexmo.com'


@responses.activate
def test_create_user_basic():
    stub(
        responses.POST,
        f'https://{host}/v1/users',
        fixture_path='users/user_basic.json',
        status_code=201,
    )

    user = users.create_user()
    assert user['id'] == 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    assert user['name'] == 'NAM-ecb938f2-13e0-40c1-9d3b-b16ebb4ef3d1'
    assert (
        user['_links']['self']['href']
        == 'https://api-us-3.vonage.com/v1/users/USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    )


@responses.activate
def test_create_user_options():
    stub(
        responses.POST,
        f'https://{host}/v1/users',
        fixture_path='users/user_options.json',
        status_code=201,
    )

    params = {
        "id": "USR-5ab17d58-b8b3-427d-ac42-c31dab7ef422",
        "name": "my_user_name",
        "image_url": "https://example.com/image.png",
        "display_name": "My User Name",
        "properties": {"custom_data": {"custom_key": "custom_value"}},
        "_links": {
            "self": {
                "href": "https://api-us-3.vonage.com/v1/users/USR-5ab17d58-b8b3-427d-ac42-c31dab7ef422"
            }
        },
        "channels": {
            "pstn": [{"number": 123457}],
            "sip": [
                {
                    "uri": "sip:4442138907@sip.example.com;transport=tls",
                    "username": "New SIP",
                    "password": "Password",
                }
            ],
            "vbc": [{"extension": "403"}],
            "websocket": [
                {
                    "uri": "wss://example.com/socket",
                    "content-type": "audio/l16;rate=16000",
                    "headers": {"customer_id": "ABC123"},
                }
            ],
            "sms": [{"number": "447700900000"}],
            "mms": [{"number": "447700900000"}],
            "whatsapp": [{"number": "447700900000"}],
            "viber": [{"number": "447700900000"}],
            "messenger": [{"id": "12345abcd"}],
        },
    }

    user = users.create_user(params)
    assert user['id'] == 'USR-5ab17d58-b8b3-427d-ac42-c31dab7ef422'
    assert user['name'] == 'my_user_name'
    assert user['display_name'] == 'My User Name'
    assert user['properties']['custom_data']['custom_key'] == 'custom_value'
    assert (
        user['_links']['self']['href']
        == 'https://api-us-3.vonage.com/v1/users/USR-5ab17d58-b8b3-427d-ac42-c31dab7ef422'
    )
    assert user['channels']['vbc'][0]['extension'] == '403'


@responses.activate
def test_create_user_400():
    stub(
        responses.POST,
        f'https://{host}/v1/users',
        fixture_path='users/user_400.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        users.create_user(params={'name': 1234})
    assert 'Input validation failure.' in str(err.value)


@responses.activate
def test_create_user_429():
    stub(
        responses.POST,
        f'https://{host}/v1/users',
        fixture_path='users/rate_limit.json',
        status_code=429,
    )

    with raises(ClientError) as err:
        users.create_user()
    assert 'You have exceeded your request limit. You can try again shortly.' in str(err.value)


@responses.activate
def test_get_user():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.GET,
        f'https://{host}/v1/users/{user_id}',
        fixture_path='users/user_basic.json',
    )

    user = users.get_user(user_id)
    assert user['name'] == 'NAM-ecb938f2-13e0-40c1-9d3b-b16ebb4ef3d1'
    assert user['properties']['custom_data'] == {}
    assert (
        user['_links']['self']['href']
        == 'https://api-us-3.vonage.com/v1/users/USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    )


@responses.activate
def test_get_user_404():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.GET,
        f'https://{host}/v1/users/{user_id}',
        status_code=404,
        fixture_path='users/user_404.json',
    )

    with raises(ClientError) as err:
        users.get_user(user_id)
    assert 'User does not exist, or you do not have access.' in str(err.value)


@responses.activate
def test_get_user_429():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.GET,
        f'https://{host}/v1/users/{user_id}',
        fixture_path='users/rate_limit.json',
        status_code=429,
    )

    with raises(ClientError) as err:
        users.get_user(user_id)
    assert 'You have exceeded your request limit. You can try again shortly.' in str(err.value)


@responses.activate
def test_update_user():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.PATCH,
        f'https://{host}/v1/users/{user_id}',
        fixture_path='users/user_updated.json',
    )

    params = {
        'name': 'updated_name',
        'channels': {
            'whatsapp': [
                {'number': '447700900000'},
            ]
        },
    }
    user = users.update_user(user_id, params)
    assert user['name'] == 'updated_name'
    assert (
        user['_links']['self']['href']
        == 'https://api-us-3.vonage.com/v1/users/USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    )
    assert user['channels']['whatsapp'][0]['number'] == '447700900000'


@responses.activate
def test_update_user_400():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.PATCH,
        f'https://{host}/v1/users/{user_id}',
        fixture_path='users/user_400.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        users.update_user(user_id, params={'name': 1234})
    assert 'Input validation failure.' in str(err.value)


@responses.activate
def test_update_user_404():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.PATCH,
        f'https://{host}/v1/users/{user_id}',
        status_code=404,
        fixture_path='users/user_404.json',
    )

    with raises(ClientError) as err:
        users.update_user(user_id, params={'name': 'updated_user_name'})
    assert 'User does not exist, or you do not have access.' in str(err.value)


@responses.activate
def test_update_user_429():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.PATCH,
        f'https://{host}/v1/users/{user_id}',
        fixture_path='users/rate_limit.json',
        status_code=429,
    )

    with raises(ClientError) as err:
        users.update_user(user_id, params={'name': 'updated_user_name'})
    assert 'You have exceeded your request limit. You can try again shortly.' in str(err.value)


@responses.activate
def test_delete_user():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.DELETE,
        f'https://{host}/v1/users/{user_id}',
        status_code=204,
        fixture_path='no_content.json',
    )

    response = users.delete_user(user_id)
    assert response == None


@responses.activate
def test_delete_user_404():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.DELETE,
        f'https://{host}/v1/users/{user_id}',
        status_code=404,
        fixture_path='users/user_404.json',
    )

    with raises(ClientError) as err:
        users.delete_user(user_id)
    assert 'User does not exist, or you do not have access.' in str(err.value)


@responses.activate
def test_delete_user_429():
    user_id = 'USR-d3cc6a55-aa7b-4916-8244-2fedb554afd5'
    stub(
        responses.DELETE,
        f'https://{host}/v1/users/{user_id}',
        fixture_path='users/rate_limit.json',
        status_code=429,
    )

    with raises(ClientError) as err:
        users.delete_user(user_id)
    assert 'You have exceeded your request limit. You can try again shortly.' in str(err.value)



class Error(Exception):
    pass


class ClientError(Error):
    pass


class ServerError(Error):
    pass


class AuthenticationError(ClientError):
    pass


class CallbackRequiredError(Error):
    """Indicates a callback is required but was not present."""


class MessagesError(Error):
    """
    Indicates an error related to the Messages class which calls the Vonage Messages API.
    """


class PricingTypeError(Error):
    """A pricing type was specified that is not allowed."""


class RedactError(Error):
    """Error related to the Redact class or Redact API."""


class InvalidAuthenticationTypeError(Error):
    """An authentication method was specified that is not allowed"""


class MeetingsError(ClientError):
    """An error related to the Meetings class which calls the Vonage Meetings API."""


class Verify2Error(ClientError):
    """An error relating to the Verify (V2) API."""


class SubaccountsError(ClientError):
    """An error relating to the Subaccounts API."""


class ProactiveConnectError(ClientError):
    """An error relating to the Proactive Connect API."""


class VideoError(ClientError):
    """An error relating to the Video API."""


class UsersError(ClientError):
    """An error relating to the Users API."""


class InvalidRoleError(ClientError):
    """The specified role was invalid."""


class TokenExpiryError(ClientError):
    """The specified token expiry time was invalid."""


class SipError(ClientError):
    """Error related to usage of SIP calls."""


class NumberVerificationError(ClientError):
    """An error relating to the Number Verification API."""



import os.path
import re

import pytest

from urllib.parse import urlparse, parse_qs

import responses


def request_body():
    return responses.calls[0].request.body


def request_query():
    return urlparse(responses.calls[0].request.url).query


def request_params():
    """Obtain the query params, as a dict."""
    return parse_qs(request_query())


def request_headers():
    return responses.calls[0].request.headers


def request_user_agent():
    return responses.calls[0].request.headers["User-Agent"]


def request_authorization():
    return responses.calls[0].request.headers["Authorization"].decode("utf-8")


def request_content_type():
    return responses.calls[0].request.headers["Content-Type"]


def stub(method, url, fixture_path=None, status_code=200):
    body = load_fixture(fixture_path) if fixture_path else '{"key":"value"}'
    responses.add(method, url, body=body, status=status_code, content_type="application/json")


def stub_bytes(method, url, body):
    responses.add(method, url, body, status=200)


def assert_re(pattern, string):
    __tracebackhide__ = True
    if not re.search(pattern, string):
        pytest.fail(f"Cannot find pattern {repr(pattern)} in {repr(string)}")


def assert_basic_auth():
    params = request_params()
    assert "api_key" not in params
    assert "api_secret" not in params
    assert request_headers()["Authorization"] == "Basic bmV4bW8tYXBpLWtleTpuZXhtby1hcGktc2VjcmV0"


def load_fixture(fixture_path):
    return open(os.path.join(os.path.dirname(__file__), "data", fixture_path)).read()



from urllib.parse import urlparse
from vonage_jwt.verify_jwt import verify_signature


class Voice:
    auth_type = 'jwt'

    def __init__(self, client):
        self._client = client

    # Creates a new call session
    def create_call(self, params, **kwargs):
        """
        Adding Random From Number Feature for the Voice API,
        if set to `True`, the from number will be randomly selected
        from the pool of numbers available to the application making
        the call.

        :param params is a dictionary that holds the 'from' and 'random_from_number'

        """
        if not params:
            params = kwargs

            key = 'from'
            if key not in params:
                params['random_from_number'] = True

        return self._client.post(
            self._client.api_host(), "/v1/calls", params or kwargs, auth_type=Voice.auth_type
        )

    # Get call history paginated. Pass start and end dates to filter the retrieved information
    def get_calls(self, params=None, **kwargs):
        return self._client.get(
            self._client.api_host(), "/v1/calls", params or kwargs, auth_type=Voice.auth_type
        )

    # Get a single call record by identifier
    def get_call(self, uuid):
        return self._client.get(
            self._client.api_host(), f"/v1/calls/{uuid}", auth_type=Voice.auth_type
        )

    # Update call data using custom ncco
    def update_call(self, uuid, params=None, **kwargs):
        return self._client.put(
            self._client.api_host(),
            f"/v1/calls/{uuid}",
            params or kwargs,
            auth_type=Voice.auth_type,
        )

    # Plays audio streaming into call in progress - stream_url parameter is required
    def send_audio(self, uuid, params=None, **kwargs):
        return self._client.put(
            self._client.api_host(),
            f"/v1/calls/{uuid}/stream",
            params or kwargs,
            auth_type=Voice.auth_type,
        )

    # Play an speech into specified call - text parameter (text to speech) is required
    def send_speech(self, uuid, params=None, **kwargs):
        return self._client.put(
            self._client.api_host(),
            f"/v1/calls/{uuid}/talk",
            params or kwargs,
            auth_type=Voice.auth_type,
        )

    # plays DTMF tones into the specified call
    def send_dtmf(self, uuid, params=None, **kwargs):
        return self._client.put(
            self._client.api_host(),
            f"/v1/calls/{uuid}/dtmf",
            params or kwargs,
            auth_type=Voice.auth_type,
        )

    # Stops audio recently played into specified call
    def stop_audio(self, uuid):
        return self._client.delete(
            self._client.api_host(), f"/v1/calls/{uuid}/stream", auth_type=Voice.auth_type
        )

    # Stop a speech recently played into specified call
    def stop_speech(self, uuid):
        return self._client.delete(
            self._client.api_host(), f"/v1/calls/{uuid}/talk", auth_type=Voice.auth_type
        )

    def get_recording(self, url):
        hostname = urlparse(url).hostname
        headers = self._client.headers
        headers['Authorization'] = self._client._create_jwt_auth_string()
        return self._client.parse(hostname, self._client.session.get(url, headers=headers))

    def verify_signature(self, token: str, signature: str) -> bool:
        return verify_signature(token, signature)



from util import *
from vonage.errors import CallbackRequiredError


@responses.activate
def test_get_basic_number_insight(number_insight, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/ni/basic/json")

    assert isinstance(number_insight.get_basic_number_insight(number="447525856424"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_query()


@responses.activate
def test_get_standard_number_insight(number_insight, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/ni/standard/json")

    assert isinstance(number_insight.get_standard_number_insight(number="447525856424"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_query()


@responses.activate
def test_get_advanced_number_insight(number_insight, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/ni/advanced/json")

    assert isinstance(number_insight.get_advanced_number_insight(number="447525856424"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_query()


@responses.activate
def test_get_async_advanced_number_insight(number_insight, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/ni/advanced/async/json")

    params = {"number": "447525856424", "callback": "https://example.com"}

    assert isinstance(number_insight.get_async_advanced_number_insight(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_query()
    assert "callback=https%3A%2F%2Fexample.com" in request_query()


def test_callback_required_error_async_advanced_number_insight(number_insight, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/ni/advanced/async/json")

    params = {"number": "447525856424", "callback": ""}

    with pytest.raises(CallbackRequiredError):
        number_insight.get_async_advanced_number_insight(params)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vonage import Client

from .errors import (
    InvalidRoleError,
    TokenExpiryError,
    SipError,
    VideoError,
)

import re
from time import time
from uuid import uuid4


class Video:
    auth_type = 'jwt'
    archive_mode_values = {'manual', 'always'}
    media_mode_values = {'routed', 'relayed'}
    token_roles = {'subscriber', 'publisher', 'publisheronly', 'moderator'}

    def __init__(self, client: Client):
        self._client = client

    def create_session(self, session_options: dict = None):
        if session_options is None:
            session_options = {}

        params = {'archiveMode': 'manual', 'p2p.preference': 'disabled', 'location': None}
        if (
            'archive_mode' in session_options
            and session_options['archive_mode'] not in Video.archive_mode_values
        ):
            raise VideoError(
                f'Invalid archive_mode value. Must be one of {Video.archive_mode_values}.'
            )
        elif 'archive_mode' in session_options:
            params['archiveMode'] = session_options['archive_mode']
        if (
            'media_mode' in session_options
            and session_options['media_mode'] not in Video.media_mode_values
        ):
            raise VideoError(f'Invalid media_mode value. Must be one of {Video.media_mode_values}.')
        elif 'media_mode' in session_options:
            if session_options['media_mode'] == 'routed':
                params['p2p.preference'] = 'disabled'
            elif session_options['media_mode'] == 'relayed':
                if params['archiveMode'] == 'always':
                    raise VideoError(
                        'Invalid combination: cannot specify "archive_mode": "always" and "media_mode": "relayed".'
                    )
                else:
                    params['p2p.preference'] = 'enabled'
        if 'location' in session_options:
            params['location'] = session_options['location']

        session = self._client.post(
            self._client.video_host(),
            '/session/create',
            params,
            auth_type=Video.auth_type,
            body_is_json=False,
        )[0]

        media_mode = self.get_media_mode(params['p2p.preference'])
        session_info = {
            'session_id': session['session_id'],
            'archive_mode': params['archiveMode'],
            'media_mode': media_mode,
            'location': params['location'],
        }

        return session_info

    def get_media_mode(self, p2p_preference):
        if p2p_preference == 'disabled':
            return 'routed'
        elif p2p_preference == 'enabled':
            return 'relayed'

    def get_stream(self, session_id, stream_id):
        return self._client.get(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/session/{session_id}/stream/{stream_id}',
            auth_type=Video.auth_type,
        )

    def list_streams(self, session_id):
        return self._client.get(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/session/{session_id}/stream',
            auth_type=Video.auth_type,
        )

    def set_stream_layout(self, session_id, items):
        return self._client.put(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/session/{session_id}/stream',
            items,
            auth_type=Video.auth_type,
        )

    def send_signal(self, session_id, type, data, connection_id=None):
        if connection_id:
            request_uri = f'/v2/project/{self._client.application_id}/session/{session_id}/connection/{connection_id}/signal'
        else:
            request_uri = f'/v2/project/{self._client.application_id}/session/{session_id}/signal'

        params = {'type': type, 'data': data}

        return self._client.post(
            self._client.video_host(), request_uri, params, auth_type=Video.auth_type
        )

    def disconnect_client(self, session_id, connection_id):
        return self._client.delete(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/session/{session_id}/connection/{connection_id}',
            auth_type=Video.auth_type,
        )

    def mute_stream(self, session_id, stream_id):
        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/session/{session_id}/stream/{stream_id}/mute',
            params=None,
            auth_type=Video.auth_type,
        )

    def mute_all_streams(self, session_id, active=True, excluded_stream_ids: list = []):
        params = {'active': active, 'excludedStreamIds': excluded_stream_ids}

        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/session/{session_id}/mute',
            params,
            auth_type=Video.auth_type,
        )

    def disable_mute_all_streams(self, session_id, excluded_stream_ids: list = []):
        return self.mute_all_streams(
            session_id, active=False, excluded_stream_ids=excluded_stream_ids
        )

    def list_archives(self, filter_params=None, **filter_kwargs):
        return self._client.get(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive',
            filter_params or filter_kwargs,
            auth_type=Video.auth_type,
        )

    def create_archive(self, params=None, **kwargs):
        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive',
            params or kwargs,
            auth_type=Video.auth_type,
        )

    def get_archive(self, archive_id):
        return self._client.get(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive/{archive_id}',
            auth_type=Video.auth_type,
        )

    def delete_archive(self, archive_id):
        return self._client.get(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive/{archive_id}',
            auth_type=Video.auth_type,
        )

    def add_stream_to_archive(self, archive_id, stream_id, has_audio=True, has_video=True):
        params = {'addStream': stream_id, 'hasAudio': has_audio, 'hasvideo': has_video}

        return self._client.patch(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive/{archive_id}/streams',
            params,
            auth_type=Video.auth_type,
        )

    def remove_stream_from_archive(self, archive_id, stream_id):
        params = {'removeStream': stream_id}

        return self._client.patch(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive/{archive_id}/streams',
            params,
            auth_type=Video.auth_type,
        )

    def stop_archive(self, archive_id):
        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive/{archive_id}/stop',
            params=None,
            auth_type=Video.auth_type,
        )

    def change_archive_layout(self, archive_id, params=None, **kwargs):
        return self._client.put(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/archive/{archive_id}/layout',
            params or kwargs,
            auth_type=Video.auth_type,
        )

    def create_sip_call(self, session_id: str, token: str, sip: dict):
        if 'uri' not in sip:
            raise SipError('You must specify a uri when creating a SIP call.')

        params = {'sessionId': session_id, 'token': token, 'sip': sip}
        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/dial',
            params,
            auth_type=Video.auth_type,
        )

    def play_dtmf(self, session_id: str, digits: str, connection_id: str = None):
        if not re.search('^[0-9*#p]+$', digits):
            raise VideoError('Only digits 0-9, *, #, and "p" are allowed.')

        params = {'digits': digits}

        if connection_id is not None:
            return self._client.post(
                self._client.video_host(),
                f'/v2/project/{self._client.application_id}/session/{session_id}/connection/{connection_id}/play-dtmf',
                params,
                auth_type=Video.auth_type,
            )

        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/session/{session_id}/play-dtmf',
            params,
            auth_type=Video.auth_type,
        )

    def list_broadcasts(self, offset: int = None, count: int = None, session_id: str = None):
        if offset is not None and (type(offset) != int or offset < 0):
            raise VideoError('Offset must be an int >= 0.')
        if count is not None and (type(count) != int or count < 0 or count > 1000):
            raise VideoError('Count must be an int between 0 and 1000.')

        params = {'offset': str(offset), 'count': str(count), 'sessionId': session_id}

        return self._client.get(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/broadcast',
            params,
            auth_type=Video.auth_type,
        )

    def start_broadcast(self, params: dict):
        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/broadcast',
            params,
            auth_type=Video.auth_type,
        )

    def get_broadcast(self, broadcast_id: str):
        return self._client.get(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/broadcast/{broadcast_id}',
            auth_type=Video.auth_type,
        )

    def stop_broadcast(self, broadcast_id: str):
        return self._client.post(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/broadcast/{broadcast_id}',
            params={},
            auth_type=Video.auth_type,
        )

    def change_broadcast_layout(self, broadcast_id: str, params: dict):
        return self._client.put(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/broadcast/{broadcast_id}/layout',
            params=params,
            auth_type=Video.auth_type,
        )

    def add_stream_to_broadcast(
        self, broadcast_id: str, stream_id: str, has_audio=True, has_video=True
    ):
        params = {'addStream': stream_id, 'hasAudio': has_audio, 'hasvideo': has_video}

        return self._client.patch(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/broadcast/{broadcast_id}/streams',
            params,
            auth_type=Video.auth_type,
        )

    def remove_stream_from_broadcast(self, broadcast_id: str, stream_id: str):
        params = {'removeStream': stream_id}

        return self._client.patch(
            self._client.video_host(),
            f'/v2/project/{self._client.application_id}/broadcast/{broadcast_id}/streams',
            params,
            auth_type=Video.auth_type,
        )

    def generate_client_token(self, session_id, token_options={}):
        now = int(time())
        claims = {
            'scope': 'session.connect',
            'session_id': session_id,
            'role': 'publisher',
            'initial_layout_class_list': '',
            'jti': str(uuid4()),
            'iat': now,
        }
        if 'role' in token_options:
            claims['role'] = token_options['role']
        if 'data' in token_options:
            claims['data'] = token_options['data']
        if 'initialLayoutClassList' in token_options:
            claims['initial_layout_class_list'] = token_options['initialLayoutClassList']
        if 'expireTime' in token_options and token_options['expireTime'] > now:
            claims['exp'] = token_options['expireTime']
        if 'jti' in token_options:
            claims['jti'] = token_options['jti']
        if 'iat' in token_options:
            claims['iat'] = token_options['iat']
        if 'subject' in token_options:
            claims['subject'] = token_options['subject']
        if 'acl' in token_options:
            claims['acl'] = token_options['acl']

        self.validate_client_token_options(claims)
        self._client.auth(claims)
        return self._client._generate_application_jwt()

    def validate_client_token_options(self, claims):
        now = int(time())
        if claims['role'] not in Video.token_roles:
            raise InvalidRoleError(
                f'Invalid role specified for the client token. Valid values are: {Video.token_roles}'
            )
        if 'exp' in claims and claims['exp'] > now + 3600 * 24 * 30:
            raise TokenExpiryError('Token expiry date must be less than 30 days from now.')



class ShortCodes:
    auth_type = 'params'
    defaults = {'auth_type': auth_type, 'body_is_json': False}

    def __init__(self, client):
        self._client = client

    def send_2fa_message(self, params=None, **kwargs):
        return self._client.post(
            self._client.host(), "/sc/us/2fa/json", params or kwargs, **ShortCodes.defaults
        )

    def send_event_alert_message(self, params=None, **kwargs):
        return self._client.post(
            self._client.host(), "/sc/us/alert/json", params or kwargs, **ShortCodes.defaults
        )

    def send_marketing_message(self, params=None, **kwargs):
        return self._client.post(
            self._client.host(), "/sc/us/marketing/json", params or kwargs, **ShortCodes.defaults
        )

    def get_event_alert_numbers(self):
        return self._client.get(
            self._client.host(), "/sc/us/alert/opt-in/query/json", auth_type=ShortCodes.auth_type
        )

    def resubscribe_event_alert_number(self, params=None, **kwargs):
        return self._client.post(
            self._client.host(),
            "/sc/us/alert/opt-in/manage/json",
            params or kwargs,
            **ShortCodes.defaults,
        )



from pydantic import BaseModel, HttpUrl, AnyUrl, constr, field_serializer, Field
from typing import Dict
from typing_extensions import Annotated, Literal


class ConnectEndpoints:
    class Endpoint(BaseModel):
        type: Literal['phone', 'app', 'websocket', 'sip', 'vbc'] = None

    class PhoneEndpoint(Endpoint):
        type: Literal['phone'] = 'phone'

        number: constr(pattern=r'^[1-9]\d{6,14}$')
        dtmfAnswer: constr(pattern='^[0-9*#p]+$') = None
        onAnswer: Dict[str, HttpUrl] = None

        @field_serializer('onAnswer')
        def serialize_dt(self, oa: Dict[str, HttpUrl], _info):
            if oa is None:
                return oa

            return {k: str(v) for k, v in oa.items()}

    class AppEndpoint(Endpoint):
        type: Literal['app'] = 'app'
        user: str

    class WebsocketEndpoint(Endpoint):
        type: Literal['websocket'] = 'websocket'

        uri: AnyUrl
        contentType: Annotated[
            Literal['audio/l16;rate=16000', 'audio/l16;rate=8000'],
            Field(serialization_alias='content-type'),
        ]
        headers: dict = None

        @field_serializer('uri')
        def serialize_uri(self, uri: AnyUrl, _info):
            return str(uri)

    class SipEndpoint(Endpoint):
        type: Literal['sip'] = 'sip'
        uri: str
        headers: dict = None

    class VbcEndpoint(Endpoint):
        type: Literal['vbc'] = 'vbc'
        extension: str

    @classmethod
    def create_endpoint_model_from_dict(cls, d) -> Endpoint:
        if d['type'] == 'phone':
            return cls.PhoneEndpoint.model_validate(d)
        elif d['type'] == 'app':
            return cls.AppEndpoint.model_validate(d)
        elif d['type'] == 'websocket':
            return cls.WebsocketEndpoint.model_validate(d)
        elif d['type'] == 'sip':
            return cls.WebsocketEndpoint.model_validate(d)
        elif d['type'] == 'vbc':
            return cls.WebsocketEndpoint.model_validate(d)
        else:
            raise ValueError(
                'Invalid "type" specified for endpoint object. Cannot create a ConnectEndpoints.Endpoint model.'
            )
